(* unification index of clauses for inst/res implementation *)

open Lib
open Options
open Statistics
open Logic_interface

(*----- debug modifiable part-----*)

let dbg_flag = false

type dbg_gr =
  | D_trace

let dbg_gr_to_str = function
  | D_trace -> "trace"

let dbg_groups =
  [
    D_trace;
 ]

let module_name = "clause_unif_index"


(*----- debug fixed part --------*)

let () = dbg_flag_msg dbg_flag module_name

let dbg group str_lazy =
  Lib.dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy

let dbg_env group f =
  Lib.dbg_env_set dbg_flag dbg_groups group f

(*----- debug -----*)

(*---------------------------------------------------------*)
(* choose an appropriate unificiation index implementation *)
(*---------------------------------------------------------*)

 module UnifIndex = UnifIndexDiscrTree 

(* let ()= out_str "\n\n !!!!!! clauseUnifIndex module UnifIndex = UnifIndexDebug !!!!!!!!!!\n\n"*)

(* module UnifIndex = UnifIndexDebug *)

(* module UnifIndex = UnifIndexMap *)
(* module UnifIndex = UnifIndexDebug *)

type t = {
  mutable unif_index : (clause UnifIndex.t) ref;
}

(* create unification index: i/r *)
let create () =
{
  unif_index = ref (UnifIndex.create ());
}

(* clear unification index: i/r, like create *)
let clear ui = UnifIndex.clear !(ui.unif_index)

(* add clause with given selection literal: i/r *)
let add_clause_with_sel ui sel_lit main_clause =
  dbg D_trace (lazy ("Add: lit: "^(Term.to_string sel_lit)^", clause: "^(Clause.to_string main_clause)));
  UnifIndex.add_elem_to_lit !(ui.unif_index) sel_lit main_clause

(* remove clause with given selection literal: i/r *)
(* Throws Not_found if sel_lit not found in unif index *)
let elim_clause_with_sel ui sel_lit main_clause =
  dbg D_trace (lazy ("Del: lit: "^(Term.to_string sel_lit)^", clause: "^(Clause.to_string main_clause)));
  UnifIndex.elim_elem_from_lit !(ui.unif_index) sel_lit main_clause

(* eliminates all clauses indexed by lit from unif_index and returns *)
(* the eliminated clause list: i *)

let eliminate_lit ui lit =
  let removed_clauses = UnifIndex.eliminate_lit !(ui.unif_index) lit in
  dbg D_trace (lazy ("Elim: lit: "^(Term.to_string lit)^", clauses: "^(Clause.clause_list_to_string removed_clauses)));
  (* return the list *)
  removed_clauses

let print_unif_cand unif_cand =
  let cl_list_to_string clause_list =
    let p_cl str clause = str^","^(Clause.to_string clause) in
    match clause_list with
    | hd::tl -> (List.fold_left p_cl ("["^(Clause.to_string hd)) tl)^"]"
    | [] -> "[]"
  in
  let print_one_cand str (lit, clause_list) = str^(Term.to_string lit)^": "^(cl_list_to_string clause_list)^"; " in
  (List.fold_left print_one_cand "[ " unif_cand)^"]"

(* get unification candidates for given selection literals, i/r *)
let get_unif_candidates ui sel_lit =
  dbg D_trace (lazy ("Unif: lit: "^(Term.to_string sel_lit)));
  let candidates = UnifIndex.get_unif_candidates !(ui.unif_index) sel_lit in
  dbg D_trace (lazy ("Cand: lit: "^(Term.to_string sel_lit)^", res: "^(print_unif_cand candidates)));
  candidates


(* check whether given literal is in the unif index *)
let in_unif_index ui lit = UnifIndex.in_unif_index !(ui.unif_index) lit
