open Lib
open Options
open Logic_interface
open Proof_search_loop
open Problem_properties

exception Schedule_Terminated

(*------------------*)

type result = 
  | Res_empty_clause of clause 
  | Res_resolution_sat of context * clause list 
  | Res_prop_solver_unsat 
  | Res_prop_solver_unsat_na (* unsat without assumtions *)
  | Res_instantiation_sat of context * clause list 
  | Res_unsat_multiple_cores of UnsatCore.unsat_core list

val result_handler_basic : result -> unit

val proved_str : unit -> string
val unknown_str : unit -> string
val satisfiable_str : unit -> string

(*------------------*)

type time = float param

type schedule = (named_options * time) list

(* proof_clauses/finite_model_clauses may be simplified and replaced during proof/model search *)

type schedule_clauses = 
    {

(* clauses used for proof search *)
     mutable proof_clauses : clause list; 

(* clauses used for finite model search  (equality may be missing) *)
     mutable finite_model_clauses : clause list;  

(* clauses removed by sem_filter; needed for model building*)
     mutable filtered_out_clauses : clause list;  

   }

val schedule_run : 
    ?prover_functions_param:(Proof_search_loop.prover_functions Lib.param) ->  schedule_clauses -> schedule -> result

val default_schedule : problem_properties -> schedule

(* just runs input_named_options *)
val trivial_schedule : unit -> schedule

val verification_epr_schedule_old : unit -> schedule

val verification_epr_schedule_tables : unit -> schedule 

val verification_epr_schedule : unit -> schedule

val sat_schedule : problem_properties -> schedule


