open Lib
open Logic_interface

(*-------------------------------------------------------------*)
(* Type to represent unsat core. Contains list of literals and *)
(* list of clauses. *)
(*-----------------------------------------------------------*)

type unsat_core =
  {
    uc_assumptions: lit list;
    uc_clauses: clause list;
  }

let create assumptions clauses =
  {
    uc_assumptions = assumptions;
    uc_clauses = clauses;
  }

let get_assumptions unsat_core = unsat_core.uc_assumptions
let get_clauses unsat_core = unsat_core.uc_clauses

let print unsat_core =
  let assumptions = get_assumptions unsat_core in
  let clauses = get_clauses unsat_core in
  let n_a = string_of_int (List.length assumptions) in
  let n_c = string_of_int (List.length clauses) in

  out_str ("unsat core: "^n_a^" assumptions, "^n_c^" clauses");
  out_str "assumptions:";
  out_str (Term.term_list_to_string assumptions);
  out_str "clauses:";
  out_str (Clause.clause_list_to_string clauses)
